# -*- coding: utf-8 -*-

"""The :mod:`py_fatigue.material.paris_curve` module collects all
functions related to :term:`Paris' law` definition.
The main class is ParisCurve.
"""

# Standard "from" imports
from typing import Any, Iterable, Optional, Tuple, Union

# Standard imports
import abc

import io

# import itertools
# import warnings

# Non-standard "from" imports
# from plotly.express import colors as px_colors

# Non-standard imports
import matplotlib.figure
import matplotlib.axis
import matplotlib.pyplot as plt
import numba as nb
import numpy as np

# import plotly.graph_objs as go
# import plotly.io as pio

# py_fatigue imports
from py_fatigue.utils import make_axes
from py_fatigue.material.sn_curve import (
    _set_color,
    _get_name,
    _check_param_couple_types,
    ensure_array,
)


class AbstractCrackGrowthCurve(metaclass=abc.ABCMeta):
    """Abstract Paris curve.
    Concrete subclasses should define methods:

        _ `_repr_svg_`
        - `format_name`
        - `get_knee_sif`
        - `get_knee_growth_rate`
        - `get_sif`
        - `get_growth_rate`
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # Eight is reasonable in this case.

    def __init__(
        self,
        slope: Union[int, float, list, np.ndarray],
        intercept: Union[int, float, list, np.ndarray],
        threshold: Union[int, float] = 0,
        critical: Union[int, float] = np.inf,
        environment: Optional[str] = None,
        curve: Optional[str] = None,
        norm: Optional[str] = None,
        unit_string: str = "MPa √mm",
        color: Union[str, None] = None,
    ) -> None:
        """Define crack growth rate curve (Paris' law).
        See class docstring for more information.

        Parameters
        ----------
        slope : Union[int, float, list, np.ndarray]
            Paris curve slope
        intercept : Union[int, float, list, np.ndarray]
            crack growth rate axis intercept
        threshold : Union[int, float]
            propagation threshold, below which crack growth rate is null
        critical : Union[int, float]
            critical propagation stress intensity factor, imminent failure
        environment : str, optional
            Paris' law envirnoment, by default None
        curve : str, optional
            Paris' law category, by default None
        norm : str, optional
            Paris' law norm, by default None
        unit_string : str, optional
            units, by default "MPa"
        color : str, optional
            RGBS or HEX string for color, by default None
        """
        self.color, self.id = _set_color(color)
        self.norm = norm
        self.environment = environment
        self.curve = curve
        self.name = _get_name(curve, environment, norm)
        self.__slope, self.__intercept = _check_param_couple_types(
            slope, intercept
        )
        self.__linear = len(self.slope) == 1
        self.__threshold = threshold
        self.__critical = critical
        self.__unit = unit_string

    def __str__(self) -> str:
        """Crack growth curve __str__ method

        Returns:
            str: Crack growth curve string description
        """
        return self.name

    def _repr_svg_(self):
        """SVG representation of the crack growth curve instance

        Returns:
            str: the SVG object
        """
        # def _repr_png_(self) -> None:
        # """PNG representation of the crack growth curve instance

        # Returns:
        #     None
        # """

    def __eq__(self, other: object) -> bool:
        """Compare different crack growth curves and assert that they
        contain same attributes values
        """
        if self.__class__ != other.__class__:
            return False

        def compare_exact(first, second):
            """Return whether two dicts of arrays are exactly equal"""
            if first.keys() != second.keys():
                return False
            return all(
                np.array_equal(first[key], second[key]) for key in first
            )

        return compare_exact(self.__dict__, other.__dict__)

    @property
    def linear(self) -> bool:
        """Preventing attribute modification outside of constructor

        Returns:
            bool: Is it a linear Paris' law?
        """
        return self.__linear

    @property
    def slope(self) -> np.ndarray:
        """Preventing attribute modification outside of constructor

        Returns:
            np.ndarray: Slope
        """
        return self.__slope

    @property
    def intercept(self) -> np.ndarray:
        """Preventing attribute modification outside of constructor

        Returns:
            np.ndarray: Intercept
        """
        return self.__intercept

    @property
    def threshold(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Threshold SIF
        """
        return self.__threshold

    @property
    def critical(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Critical SIF
        """
        return self.__critical

    @property
    def unit(self) -> str:
        """Preventing attribute modification outside of constructor

        Returns:
            unit: Units of measure, by default "MPa"
        """
        return self.__unit

    @abc.abstractmethod
    def format_name(self, html_format: bool = False) -> None:
        """Reformat ParisCurve name.

        Parameters
        ----------
        html_format : bool, optional
            Choose whether the name format shall be HTML or not,
            by default True
        """

    @property
    @abc.abstractmethod
    def threshold_growth_rate(self) -> float:
        """Calculates the crack growth rate at threshold, if threshold
        SIF is defined.

        Returns
        -------
        float
            threshold crack growth rate
        """

    @property
    @abc.abstractmethod
    def critical_growth_rate(self) -> float:
        """Calculates the crack growth rate at critical, if critical
        SIF is defined.

        Returns
        -------
        float
            critical crack growth rate
        """

    @abc.abstractmethod
    def get_knee_growth_rate(
        self,
        check_knee: Union[Iterable, None] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the crack growth rate at the knee, if the Paris'
        law is more than linear.

        Parameters
        ----------
        check_knee : Iterable, optional
            Iterable of SIF values to check for the knee, by default None
        significant_digits : int, optional
            Number of significant digits to round the knee to, by default 2

        Returns
        -------
        1darray
            knee crack growth rate
        """

    @abc.abstractmethod
    def get_knee_sif(
        self,
        check_knee: Union[Iterable, None] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the SIF at the knee, if the Paris' law is more
        than linear.

        Parameters
        ----------
        check_knee : iterable, optional
            SIF values to check the knee SIF against, by default None
        significant_digits : int, optional
            number of significant digits to check the knee SIF against,
            by default 2

        Returns
        -------
        np.ndarray
            knee SIF
        """

    @abc.abstractmethod
    def get_growth_rate(
        self, sif_range: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Return cycles value(s) for stress range(s).

        Parameters
        ---------
        sif_range: int or float or list or 1darray
           :term:`SIF range(s)<SIF range>` to find the corresponding
           :term:`crack growth rate(s)<Crack growth rate>` for.

        Returns
        -------
        float or 1darray
            :term:`crack growth rate(s)<Crack growth rate>` for the
            :term:`SIF range(s)<SIF range>` indicating `sif_range`.
        """

    @abc.abstractmethod
    def get_sif(
        self, growth_rate: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Return :term:`SIF range(s)<SIF range>` for the endurance(s) N.

        Parameters
        ---------
        cycles : int or float or list or 1darray
            :term:`crack growth rate(s)<Crack growth rate>` to find the
            corresponding :term:`SIF range(s)<SIF range>` for.

        Returns
        -------
        float or 1darray
            :term:`SIF range(s)<SIF range>` for corresponding
            :term:`crack growth rate(s)<Crack growth rate>`.
        """


class ParisCurve(AbstractCrackGrowthCurve):
    """Define a crack growth curve inthe form of Paris' law that can
    have an arbitrary number of slopes/intercepts as well as a
    threshold and critical stress intensity factor. For example:

    >>> #
    >>> #         ^                log da/dN - log ΔK
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                             *  .
    >>> #         │                          *  │  .
    >>> #         │                       *     │ m2
    >>> #         │                    *        │  .
    >>> #         │                 *-----------┘  .
    >>> #         │              *                 .
    >>> #         │           .*│                  .
    >>> #         │        . *  │ m1               .
    >>> #         │     .  *    │                  .
    >>> #         │  .   *------┘                  .
    >>> # (da/dN)1+    *                           .
    >>> #         │  . *                           .
    >>> #         │.   *                           .
    >>> # (da/dN)2+----|---------------------------|--------------->
    >>> #             ΔK0                         ΔKc
    >>> #                          Number of cycles

    The slope-intercept couples `((m1, log_a1), (m2, log_a2))`, and the
    endurance value `(Ne)` are the parameters necessary to fully describe this
    trilinear Paris' law. If the endurance is not set, it defaults to Inf.

    Example
    -------

    >>> from py_fatigue import ParisCurve
    >>> import numpy as np

    >>> # Define a Paris' law with 5 slopes and 5 interceps
    >>> SIF = np.linspace(1,2500, 300)
    >>> SLOPE_5 = np.array([2.88, 5.1, 8.16, 5.1, 2.88])
    >>> INTERCEPT_5 = np.array([1E-16, 1E-20, 1E-27, 1E-19, 1E-13])
    >>> THRESHOLD = 20
    >>> CRITICAL = 2000

    >>> pc = ParisCurve(slope=SLOPE_5, intercept=INTERCEPT_5,
                      threshold=THRESHOLD, critical=CRITICAL, norm="The norm",
                      environment="Environment", curve="nr. 4")
    >>> pc_.get_knee_sif()
    array([ 63.35804993, 193.9017369 , 411.50876026, 504.31594872])

    """

    def _repr_svg_(self) -> str:
        fig, ax = self.plot()
        # fig.dpi = 100
        fig.set_figheight(5)
        fig.set_figwidth(7)
        max_x = np.max(ax.lines[0].get_xdata())
        max_y = np.max(ax.lines[0].get_ydata())
        cg_unit = "[L][CYCLE]$^{-1}$"
        if "√" in self.unit:
            split_unit = self.unit.split("√")
            cg_unit = "/".join((split_unit[1], "cycle"))

        ann_1 = f"{self.name},\nSlope: {self.slope},\nIntercept:"
        ann_2 = f"{self.intercept},\nThreshold: {self.threshold:.2f},"
        ann_3 = f"{cg_unit},\nCritical: {self.critical:.2f}, {self.unit}"
        annotation = " ".join((ann_1, ann_2, ann_3))
        ax.text(
            1.5 * max_x,
            max_y,
            annotation,
            horizontalalignment="left",
            verticalalignment="top",
            color="#435580",
            bbox={
                "facecolor": "#FFF",
                "alpha": 0.95,
                "pad": 10,
                "edgecolor": "#0E76BC",
            },
            fontsize=8,
            fontweight="semibold",
        )
        if not self.linear:
            knee_zip = zip(self.get_knee_sif(), self.get_knee_growth_rate())
            for j, (knee_sif, knee_gr) in enumerate(knee_zip):
                annotation = (
                    f"Knee {j +1}: ({round(knee_sif, 2)}, {knee_gr:.2E})"
                )
                ax.plot(knee_sif, knee_gr, "o", color="#C40000")
                for item in (
                    [ax.title, ax.xaxis.label, ax.yaxis.label]
                    + ax.get_xticklabels()
                    + ax.get_yticklabels()
                ):
                    item.set_fontsize(10)
                ax.text(
                    1.1 * knee_sif,
                    self.get_growth_rate(knee_sif),
                    annotation,
                    horizontalalignment="left",  # if j % 2 == 0 else "left",
                    verticalalignment="center",
                    fontsize=8,
                    color="#C40000",
                )
        if self.critical < np.inf:
            annotation = (
                f"Critical: ({round(self.critical, 2)}, "
                + f"{self.critical_growth_rate:.2E})"
            )
            ax.plot(
                self.critical, self.critical_growth_rate, "d", color="#435580"
            )
            ax.text(
                0.9 * self.critical,
                self.critical_growth_rate,
                annotation,
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=8,
                color="#435580",
            )
        if self.threshold > 0:
            annotation = (
                f"Threshold: ({round(self.threshold, 2)}, "
                + f"{self.threshold_growth_rate:.2E})"
            )
            ax.plot(
                self.threshold,
                self.threshold_growth_rate,
                "d",
                color="#435580",
            )
            ax.text(
                1.1 * self.threshold,
                self.threshold_growth_rate,
                annotation,
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=8,
                color="#435580",
            )
        fig.tight_layout()

        output = io.BytesIO()
        fig.savefig(output, format="svg")
        data = output.getvalue()  # .encode('utf-8') doesn't change anything
        plt.close(fig)
        return data.decode("utf-8")

    def format_name(self, html_format: bool = False) -> None:
        if not html_format:
            self.name = self.name.replace("<br> ", "\n")
        else:
            self.name = self.name.replace("\n", "<br> ")

    @property
    def threshold_growth_rate(self) -> float:
        """Calculates the crack growth rate at threshold, if threshold
        SIF is defined.

        Returns
        -------
        float
            threshold crack growth rate
        """
        return self.get_growth_rate(self.threshold)

    @property
    def critical_growth_rate(self) -> float:
        """Calculates the crack growth rate at critical, if critical
        SIF is defined.

        Returns
        -------
        float
            critical crack growth rate
        """
        return self.get_growth_rate(self.critical)

    def get_knee_growth_rate(
        self,
        check_knee: Union[Iterable, None] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the crack growth rate at the knee, if the Paris'
        law is more than linear.

        Parameters
        ----------
        check_knee : Iterable, optional
            Iterable of SIF values to check for the knee, by default None
        significant_digits : int, optional
            Number of significant digits to round the knee to, by default 2

        Returns
        -------
        np.ndarray
            knee crack growth rate
        """
        knee_growth_rate = np.empty(self.intercept.size - 1, dtype=np.float64)
        if not self.linear:
            for i in range(self.intercept.size - 1):
                m_i = self.slope[i + 1] / (self.slope[i + 1] - self.slope[i])
                m_i_p_1 = self.slope[i] / (self.slope[i + 1] - self.slope[i])
                knee_growth_rate[i] = (
                    self.intercept[i] ** m_i / self.intercept[i + 1] ** m_i_p_1
                )
            if check_knee is not None:
                try:
                    iter(check_knee)  # type: ignore
                except TypeError:
                    check_knee = np.asarray([check_knee])
                else:
                    check_knee = np.asarray(check_knee)
                if len(check_knee) != len(self.slope) - 1:
                    raise ValueError(
                        f"{len(self.slope) - 1} knee points expected"
                    )
                check_knee_growth_rate = self.get_growth_rate(check_knee)
                _ = [
                    np.testing.assert_approx_equal(  # type: ignore
                        c_k_s, k_s, significant=significant_digits
                    )
                    for c_k_s, k_s in zip(
                        check_knee_growth_rate, knee_growth_rate
                    )
                ]
        else:
            if check_knee is not None:
                raise ValueError("0 knee points expected")
        return knee_growth_rate

    def get_knee_sif(
        self,
        check_knee: Union[Iterable, None] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the SIF at the knee, if the Paris' law is more
        than linear.

        Parameters
        ----------
        check_knee : iterable, optional
            SIF values to check the knee SIF against, by default None
        significant_digits : int, optional
            number of significant digits to check the knee SIF against,
            by default 2

        Returns
        -------
        np.ndarray
            knee SIF
        """
        knee_sif = np.empty(self.intercept.size - 1, dtype=np.float64)
        if not self.linear:
            if self.intercept.size > 1:
                for i in range(self.intercept.size - 1):
                    knee_sif[i] = (
                        self.intercept[i] / self.intercept[i + 1]
                    ) ** (1 / (self.slope[i + 1] - self.slope[i]))
            if check_knee is not None:
                try:
                    iter(check_knee)  # type: ignore
                except TypeError:
                    check_knee = np.asarray([check_knee])
                else:
                    check_knee = np.asarray(check_knee)
                if len(check_knee) != len(self.slope) - 1:
                    raise ValueError(
                        f"{len(self.slope) - 1} knee points expected"
                    )
                _ = [
                    np.testing.assert_approx_equal(  # type: ignore
                        c_k_s, k_s, significant=significant_digits
                    )
                    for c_k_s, k_s in zip(check_knee, knee_sif)
                ]
        else:
            if check_knee is not None:
                raise ValueError("0 knee points expected")
        return knee_sif

    @ensure_array
    def get_growth_rate(
        self, sif_range: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        return _calc_growth_rate(
            sif_range,
            self.slope,
            self.intercept,
            self.threshold,
            self.critical,
        )

    @ensure_array
    def get_sif(
        self, growth_rate: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        return _calc_sif(
            growth_rate,
            self.slope,
            self.intercept,
            self.threshold,
            self.critical,
        )

    def plot(
        self,
        sif: Optional[list] = None,
        growth_rate: Optional[list] = None,
        dataset_name: Optional[str] = None,
        dataset_color: str = "#000",
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.collections.PathCollection] = None,
        **kwargs: Any,
    ) -> Tuple[
        matplotlib.figure.Figure, matplotlib.collections.PathCollection
    ]:

        """Use matplotlib to plot the Paris' law and a da/dN vs ΔK
        history dataset.

        Example
        -------

        Use matplotlib to plot the Paris' law and a da/dN vs ΔK history

        >>> fig, ax = pc.plot()

        Parameters
        ----------
        sif : list, optional
            SIF value(s) in the dataset,
            by default None
        growth_rate : list, optional
            :term:`Crack growth rate(s)<Crack growth rate>` in the dataset,
            by default None
        dataset_name : str, optional
            history dataset name, by default None
        dataset_color : str, optional
            history dataset color, by default "#000"
        fig : matplotlib.figure.Figure, optional
            figure object, by default None
        ax : matplotlib.collections.PathCollection, optional
            axis object, by default None
        **kwargs : Any, optional
            additional keyword arguments

        Returns
        -------
        matplotlib.figure.Figure, matplotlib.collections.PathCollection
            The figure and axes.
        """
        fig, ax = make_axes(fig, ax)
        sif_plot, growth_rate_plot = _paris_curve_data_points(self)

        # plt.figure(figsize=(6, 2.5))
        ax.loglog(
            sif_plot,
            growth_rate_plot,
            label=self.name,
            color=self.color,
            **kwargs,
        )
        if sif is not None and growth_rate is not None:
            sif, growth_rate = _check_param_couple_types(sif, growth_rate)
            ax.loglog(
                sif,
                growth_rate,
                label=dataset_name,
                color=dataset_color,
                linewidth=1,
                marker="o",
                markersize=4,
            )

        plt.grid(which="both", linestyle=":")
        ax.set_xlabel(f"SIF range (ΔK), {self.unit}")
        if "√" in self.unit:
            split_unit = self.unit.split("√")
            y_unit = "/".join((split_unit[1], "cycle"))
        else:
            y_unit = "[L][CYCLE]$^{-1}$"
        ax.set_ylabel(f"Crack growth rate (da/dN), {y_unit}")

        return fig, ax


@nb.njit(
    # 'float64[::1](float64[::1], float64[::1], float64[::1])',
    fastmath=False,
    parallel=True,
)
def _calc_growth_rate(sif, slope, intercept, threshold, critical):
    # pylint: disable=not-an-iterable
    assert intercept.size > 0 and intercept.size == slope.size
    assert np.min(sif) >= 0
    assert 0 <= threshold < critical <= np.inf

    knees_sif = np.empty(intercept.size - 1, dtype=np.float64)
    if intercept.size > 1:
        for i in nb.prange(intercept.size - 1):
            knees_sif[i] = (intercept[i] / intercept[i + 1]) ** (
                1 / (slope[i + 1] - slope[i])
            )
    knees_sif = np.hstack(
        (
            np.array([0.9999999999 * threshold]),
            knees_sif,
            np.array([critical / 0.9999999999]),
        )
    )
    e_msg = (
        "Knee(s) not in between threshold and critical SIF."
        + "\nCheck the definitions of slope, intercept, threshold, critical."
    )
    assert np.all(np.diff(knees_sif) > 0), e_msg
    # print("knees_sif:", knees_sif)
    idx = np.digitize(sif, knees_sif, right=False) - 1
    # print("idx:", idx)
    the_growth_rate = np.empty(sif.size, dtype=np.float64)
    for i in nb.prange(sif.size):
        if sif[i] < threshold:  # below threshold
            the_growth_rate[i] = 0
            continue
        if sif[i] > critical:  # above threshold
            the_growth_rate[i] = np.inf
            continue
        the_growth_rate[i] = intercept[idx[i]] * sif[i] ** slope[idx[i]]

    return the_growth_rate


@nb.njit(
    # 'float64[::1](float64[::1], float64[::1], float64[::1])',
    fastmath=False,
    parallel=True,
)
def _calc_sif(growth_rate, slope, intercept, threshold, critical):
    # pylint: disable=not-an-iterable
    assert intercept.size > 0 and intercept.size == slope.size
    assert np.min(growth_rate) >= 0
    assert 0 <= threshold < critical <= np.inf

    knees_growth_rate = np.empty(intercept.size - 1, dtype=np.float64)
    if intercept.size > 1:
        for i in nb.prange(intercept.size - 1):
            m_i = slope[i + 1] / (slope[i + 1] - slope[i])
            m_i_p_1 = slope[i] / (slope[i + 1] - slope[i])
            knees_growth_rate[i] = (
                intercept[i] ** m_i / intercept[i + 1] ** m_i_p_1
            )

    knees_growth_rate = np.hstack(
        (
            np.array([intercept[0] * (0.9999999999 * threshold) ** slope[0]]),
            knees_growth_rate,
            np.array([intercept[-1] * (critical / 0.9999999999) ** slope[-1]]),
        )
    )
    e_msg = (
        "Knee(s) not in between threshold and critical SIF."
        + "\nCheck the definitions of slope, intercept, threshold, critical."
    )
    assert np.all(np.diff(knees_growth_rate) > 0), e_msg
    # print("knees_growth_rate:", knees_growth_rate)
    idx = np.digitize(growth_rate, knees_growth_rate, right=False) - 1
    # print("idx:", idx)
    the_sif = np.empty(growth_rate.size, dtype=np.float64)
    nr_knees = intercept.size - 1
    for i in nb.prange(growth_rate.size):
        if idx[i] <= 0:  # below threshold
            the_sif[i] = threshold
            continue
        if idx[i] >= nr_knees:  # above threshold
            the_sif[i] = critical
            continue
        the_sif[i] = (growth_rate[i] / intercept[idx[i]]) ** (
            1 / slope[idx[i]]
        )

    return the_sif


def _paris_curve_data_points(pc: ParisCurve) -> tuple:
    """Draw Paris' curve data points, including possible knees and
    threshold/critical SIF.

    Parameters
    ----------
    pc : ParisCurve
        Instance of :class:`py_fatigue.ParisCurve`

    Returns
    -------
    tuple
        tuple of arrays
    """

    # if pc.threshold == 0 and pc.critical == np.inf:
    threshold: Union[int, float] = 1.0
    critical: Union[int, float] = 1000.0
    if pc.threshold > 0 and pc.critical < np.inf:
        threshold = pc.threshold
        critical = pc.critical
    if pc.threshold == 0 and pc.critical < np.inf:
        threshold = pc.critical / 100
        critical = pc.critical
    if pc.threshold > 0 and pc.critical == np.inf:
        threshold = pc.threshold
        critical = pc.threshold * 100

    sif_plot = np.logspace(
        np.log10(threshold), np.log10(0.999999999999 * critical), 20
    )  # 0.999999999999 to avoid rounding errors

    if not pc.linear:
        sif_plot = np.sort(np.append(sif_plot, np.array(pc.get_knee_sif())))
    growth_rate_plot = pc.get_growth_rate(sif_plot)
    if pc.threshold > 0:
        growth_rate_plot = np.hstack(
            [growth_rate_plot[0] / 10, growth_rate_plot]
        )
        sif_plot = np.hstack([pc.threshold, sif_plot])
    if pc.critical < np.inf:
        growth_rate_plot = np.hstack(
            [growth_rate_plot, growth_rate_plot[-1] * 10]
        )
        # Again, 0.999999999999 to avoid rounding errors
        sif_plot = np.hstack([sif_plot, 0.999999999999 * pc.critical])
    return sif_plot, growth_rate_plot



# Packages from non-standard libraries
from typing import Dict, Optional

import matplotlib
import numba as nb
import numpy as np
from pydantic import PositiveFloat, validator
from pydantic.dataclasses import dataclass

# Local imports
import py_fatigue.utils as pfu
from py_fatigue.geometry import AbstractCrackGeometry


@dataclass(repr=False)
class HollowCylinder(AbstractCrackGeometry):
    """Class to initalise and store the crack geometry data on a
    hollow cylindrical section.

    The crack is assumed to maintain its width to depth ratio.

    >>> # Example drawing of a hollow cylinder with external crack
    >>> #        ┌────────────────────────────────────────┐
    >>> #        │              ▄▄▄▀▀▀▀▀▀▀▚▄▄▖            │
    >>> #        │          ▗▄▀▀             ▝▀▚▄         │
    >>> #        │        ▄▀▘                    ▀▖       │
    >>> #        │       ▞    crack front         ▝▀▖     │
    >>> #        │     ▗▀     │      ▄▄▄▖           ▝▖    │
    >>> #        │    ▗███▄▖<─┘  ▗▟▀▀   ▀▀▚▖         ▐    │
    >>> #        │    ▟█████▖   ▟▘         ▀▙         ▌   │
    >>> #        │2c─>███████   ▌       r_i ▐         ▐   │
    >>> #        │    ███████   ▌     ↑<───>▐         ▐   │
    >>> #        │    ▜█████▘   ▜▖    │    ▄▛         ▌   │
    >>> #        │   │▝███▀▘│    ▝▜▄▄ │ ▄▄▞▘         ▐    │
    >>> #        │   │ ▝▄   │        ▀│▀▘           ▗▘    │
    >>> #        │   │   ▚  │      r_o│           ▗▄▘     │
    >>> #        │   │    ▚▄│         │          ▗▘       │
    >>> #        │   │   a  ▀▚▄▖      ↓       ▄▞▀▘        │
    >>> #        │   │<────>│  ▝▀▀▚▄▄▄▄▄▄▄▄▀▀▀            │
    >>> #        └────────────────────────────────────────┘
    >>> #

    Parameters
    ----------
    initial_depth : float
        The initial depth of the crack.
    outer_diameter : float
        The outer diameter of the part.
    thickness : float
        The thickness of the part.
    height : float
        The height of the part.
    crack_position : float
        The position of the crack on the part.
    width_to_depth_ratio : float
        The ratio of the width to the depth of the crack.
    """

    outer_diameter: PositiveFloat
    thickness: PositiveFloat
    height: PositiveFloat
    crack_position: str
    width_to_depth_ratio: PositiveFloat = 2.0

    _id = property(
        lambda self: "HOL_CYL_00"
        if self.crack_position == "internal"
        else "HOL_CYL_01"
    )

    @validator("thickness")
    @classmethod
    def thickness_validator(
        cls, v: PositiveFloat, values: dict
    ) -> PositiveFloat:
        """Validate the thickness of the part."""
        if v >= values["outer_diameter"] / 2:
            e_msg = "thickness is greater than or equal to the outer radius"
            raise ValueError(e_msg)
        if v <= values["initial_depth"]:
            e_msg = "thickness is less than or equal to the initial depth"
            raise ValueError(e_msg)
        return v

    @validator("crack_position")
    @classmethod
    def crack_position_validator(cls, v: str) -> str:
        """Validate the crack position string."""
        if v not in ["internal", "external"]:
            raise ValueError(
                "crack position must be either 'internal' or 'external'"
            )
        return v

    @property
    def geometry_factor(self) -> np.ndarray:
        """Get the geometric factor. This should be a function of the crack
        size. The default value is 1.0."""
        return np.empty(0)

    def __str__(self) -> str:
        return (
            super().__str__()[:-3]
            + f",\n  outer_diameter={self.outer_diameter}"
            + f",\n  thickness={self.thickness}"
            + f",\n  height={self.height}"
            + f",\n  width_to_depth_ratio={self.width_to_depth_ratio}"
            + f",\n  crack_position={self.crack_position},\n)"
        )

    def plot(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.collections.PathCollection] = None,
        **kwargs: dict,
    ) -> tuple:  # pragma: no cover
        """Plot the crack front on a hollow cylinder.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure to plot on.
        ax : matplotlib.collections.PathCollection, optional
            The axis to plot on.
        **kwargs : dict
            Keyword arguments to pass to the matplotlib plot function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure that was plotted on.
        ax : matplotlib.collections.PathCollection
        """

        if self.crack_position == "internal":
            raise NotImplementedError(
                "Plotting of internal crack not implemented"
            )

        # case crack_position external
        return self._plot_external_crack(fig, ax, **kwargs)

    def _plot_external_crack(
        self,
        depth: Optional[PositiveFloat] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.collections.PathCollection] = None,
        **kwargs: dict,
    ) -> tuple:
        """Plot the crack front when the crack is located on the external
        surface of the hollow cylinder.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure to plot on.
        ax : matplotlib.collections.PathCollection, optional
            The axis to plot on.
        **kwargs : dict
            Keyword arguments to pass to the matplotlib plot function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure that was plotted on.
        ax : matplotlib.collections.PathCollection
        """
        if depth is None:
            depth = self.initial_depth
        angles = np.linspace(0, 2 * np.pi, 100)
        r_o = self.outer_diameter / 2
        r_i = r_o - self.thickness

        x_o = r_o * np.cos(angles)
        y_o = r_o * np.sin(angles)
        x_i = r_i * np.cos(angles)
        y_i = r_i * np.sin(angles)

        crack_width = self.width_to_depth_ratio * depth
        c_angle = crack_width / r_o  # crack angle
        c_v = r_o * np.sin(c_angle)
        a_h = depth - r_o * (1 - np.cos(c_angle))
        r_e = (c_v**2 + a_h**2) / (2 * a_h)  # radius of crack-front arc
        # angle of crack-front arc
        beta = np.arcsin(c_v / r_e)
        theta = np.linspace(-beta, beta, 100)  # θ (crack-front angle)
        phi = np.linspace(np.pi - c_angle, np.pi + c_angle, 100)

        # crack-front arc
        x_c = r_e * np.cos(theta) - (r_o + r_e - depth)
        y_c = r_e * np.sin(theta)
        x_c = np.append(x_c, r_o * np.cos(phi))
        y_c = np.append(y_c, r_o * np.sin(phi))

        # plot
        fig, axes = pfu.make_axes(fig=fig, ax=ax)
        axes.set_aspect("equal")
        axes.plot(x_o, y_o, c="k", lw=0.5, **kwargs)
        axes.plot(x_i, y_i, c="k", lw=0.5, **kwargs)
        axes.plot(x_c, y_c, c="r", lw=1.0, **kwargs)
        return fig, axes


@dataclass(repr=False)
class Cylinder(AbstractCrackGeometry):
    """Class to initalise and store the crack geometry data on a
    cylindrical section.

    Parameters
    ----------
    diameter : float
        The outer diameter of the part.
    height : float
        The height of the part.
    """

    diameter: PositiveFloat
    height: PositiveFloat
    _id = property(lambda _: "FUL_CYL_00")

    def __post_init__(self):
        if not all((self.diameter, self.height)) > 0.0:
            raise ValueError("Diameter and height must be positive.")

    @property
    def geometry_factor(self) -> np.ndarray:
        """Get the geometric factor. This should be a function of the crack
        size. The default value is an empty."""
        return np.empty(0)

    def __str__(self) -> str:
        return (
            super().__str__()[:-3]
            + f",\n  diameter={self.diameter}"
            + f",\n  height={self.height},\n)"
        )


@nb.njit(
    nb.float64(
        nb.float64, nb.types.DictType(nb.types.unicode_type, nb.float64)
    ),
    cache=True,
    fastmath=True,
)
def f_hol_cyl_01(
    crack_depth: float,
    crack_geometry: Dict[str, float],
):
    """Calculate the crack geometry factor for a hollow cylinder with
    an external crack.

    Parameters
    ----------
    crack_depth : float
        The crack depth.
    outer_diameter : float
        The outer diameter of the part.
    thickness : float
        The thickness of the part.
    height : float
        The height of the part.

    Returns
    -------
    float
        The crack size factor.
    """
    if crack_depth >= crack_geometry["thickness"]:
        return 0.0
    outer_diameter: float = crack_geometry["outer_diameter"]
    thickness: float = crack_geometry["thickness"]
    width_to_depth_ratio: float = crack_geometry["width_to_depth_ratio"]
    crack_width: float = crack_depth * width_to_depth_ratio
    outer_radius: float = outer_diameter / 2
    inner_radius: float = outer_radius - thickness
    # raise NotImplementedError
    # crack angle
    alpha: float = crack_width / outer_radius
    c_v: float = outer_radius * np.sin(alpha)
    a_h: float = crack_depth - outer_radius * (1 - np.cos(alpha))
    # radius of crack-front arc
    crack_radius: float = (c_v**2 + a_h**2) / (2 * a_h)
    # angle of crack-front arc
    beta: float = np.arcsin(c_v / crack_radius)
    # areas
    # * circular ring
    area_0: float = np.pi * (outer_radius**2 - inner_radius**2)
    # * external circular segment
    area_1: float = (alpha - np.sin(2 * alpha) / 2) * outer_radius**2
    # internal circular segment
    area_2: float = (beta - np.sin(2 * beta) / 2) * crack_radius**2
    # * net crack area
    area_net: float = area_0 - area_1 - area_2
    # centroids of the circular segments
    x_1: float = (
        -4
        * outer_radius
        * np.sin(alpha) ** 3
        / (3 * (2 * alpha - np.sin(2 * alpha)))
    )
    x_2: float = (
        4
        * crack_radius
        * np.sin(beta) ** 3
        / (3 * (2 * beta - np.sin(2 * beta)))
        - outer_radius * np.cos(alpha)
        - crack_radius * np.cos(beta)
    )
    # location of the net section centroid
    q_x: float = -(area_1 * x_1 + area_2 * x_2) / area_net
    # moments of inertia
    # * circular ring
    i_0: float = (
        np.pi / 4 * (outer_radius**4 - inner_radius**4) + area_0 * q_x**2
    )
    # * external circular segment
    i_1: float = (outer_radius**4) / 4 * (
        alpha
        - np.sin(4 * alpha) / 4
        - (16 * np.sin(alpha) ** 6)
        / (9 * (alpha - np.sin(alpha) * np.cos(alpha)))
    ) + area_1 * (q_x - x_1) ** 2
    # * internal circular segment
    i_2: float = (
        crack_radius**4
        / 4
        * (
            beta
            - np.sin(4 * beta) / 4
            - (16 * np.sin(beta) ** 6)
            / (9 * (beta - np.sin(beta) * np.cos(beta)))
        )
        + area_2 * (q_x - x_2) ** 2
    )
    # * net area moment of inertia
    i_net: float = i_0 - i_1 - i_2
    # net section stress center
    epsilon: float = 0.1
    # * loc a
    x_a: float = -(
        outer_radius * np.cos(alpha) + q_x
    ) + epsilon * outer_radius * (1 + np.cos(alpha))
    # # * loc b
    # x_b: float = (outer_radius - q_x) - epsilon * outer_radius * \
    # (1 + np.cos(alpha))
    # net section stress geometry factors
    # geometry factors
    # # * net section stress factor on sigma_0
    # g_0: float = area_0 / area_net - area_0 * q_x * x_a / i_net
    # * net section stress factor on sigma_1
    g_1: float = (
        -np.pi
        * (outer_radius**4 - inner_radius**4)
        * x_a
        / (4 * outer_radius * i_net)
    )
    return g_1



from py_fatigue.cycle_count.cycle_count import CycleCount

from py_fatigue.material.sn_curve import SNCurve
from py_fatigue.material.crack_growth_curve import ParisCurve
from py_fatigue.damage import stress_life, crack_growth
from py_fatigue.version import __version__
from py_fatigue.cycle_count import rainflow, histogram
from py_fatigue import styling, testing, utils, geometry

__all__ = [
    "CycleCount",
    "SNCurve",
    "ParisCurve",
    "stress_life",
    "crack_growth",
    "rainflow",
    "histogram",
    "geometry",
    "styling",
    "testing",
    "utils",
    "__version__",
]



"""
The :mod:`py_fatigue.styling` module collects all the functions and
classes related with better displaying i/o operations.
"""

from typing import Optional, Type, Union

__all__ = ["TermColors", "py_fatigue_formatwarning"]


class TermColors:  # pragma: no cover
    """Color set for terminal outputs"""

    # pylint: disable=too-few-public-methods

    CEND = "\33[0m"
    CBOLD = "\33[1m"
    CITALIC = "\33[3m"
    CURL = "\33[4m"
    CBLINK = "\33[5m"
    CBLINK2 = "\33[6m"
    CSELECTED = "\33[7m"

    CBLACK = "\33[30m"
    CRED = "\33[31m"
    CGREEN = "\33[32m"
    CYELLOW = "\33[33m"
    CBLUE = "\33[34m"
    CVIOLET = "\33[35m"
    CBEIGE = "\33[36m"
    CWHITE = "\33[37m"

    CBLACKBG = "\33[40m"
    CREDBG = "\33[41m"
    CGREENBG = "\33[42m"
    CYELLOWBG = "\33[43m"
    CBLUEBG = "\33[44m"
    CVIOLETBG = "\33[45m"
    CBEIGEBG = "\33[46m"
    CWHITEBG = "\33[47m"

    CGREY = "\33[90m"
    CRED2 = "\33[91m"
    CGREEN2 = "\33[92m"
    CYELLOW2 = "\33[93m"
    CBLUE2 = "\33[94m"
    CVIOLET2 = "\33[95m"
    CBEIGE2 = "\33[96m"
    CWHITE2 = "\33[97m"

    CGREYBG = "\33[100m"
    CREDBG2 = "\33[101m"
    CGREENBG2 = "\33[102m"
    CYELLOWBG2 = "\33[103m"
    CBLUEBG2 = "\33[104m"
    CVIOLETBG2 = "\33[105m"
    CBEIGEBG2 = "\33[106m"
    CWHITEBG2 = "\33[107m"


def py_fatigue_formatwarning(
    message: Union[Warning, str],
    category: Type[Warning],
    filename: str,
    lineno: int,
    line: Optional[str] = None,
) -> str:  # pragma: no cover
    """Monkeypatch the warning messages.

    Parameters
    ----------
    message : Union[Warning, str]
        The warning message.
    category : Type[Warning]
        The warning category.
    filename : str
        The file name.
    lineno : int
        The warning line number.
    line : Optional[str], optional
        The warning line, by default None

    Returns
    -------
    str
        Monkeypatched warning message
    """
    # pylint: disable=unused-argument
    if category:
        w_msg = "".join(
            (
                f"{TermColors.CRED2}{TermColors.CBOLD}",
                f"{str(category.__name__,)}:\n{TermColors.CEND}",
                f"{TermColors.CYELLOW2}{message}{TermColors.CEND}\n",
            )
        )
        return w_msg
    return f"{TermColors.CYELLOW2}{message}{TermColors.CEND}\n"
